//3456789_123456789_123456789_123456789_123456789_123456789_123456789_123456789_
//
//
// quikMeshImg.cpp: implementation of the CquikMeshImg class.
//
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "Proj1.h"
#include "quikMeshImg.h"

#ifdef _DEBUG
#undef THIS_FILE
static char THIS_FILE[]=__FILE__;
#define new DEBUG_NEW
#endif


void CquikMeshImg::whatsup(void)
//------------------------------------------------------------------------------
{// BREAKPOINT HERE to diagnose warnings through the call stack.
	MessageBeep(MB_ICONASTERISK);
}

CquikMeshImg::CquikMeshImg()							
//------------------------------------------------------------------------------
// Constructor, initializer
{
	m_pTimg = NULL;							// no texture image,
	m_pNorm = NULL;							// No normal vector array exists,
	m_pTexy = NULL;							// No texture coord array exists,
	m_pVert = NULL;							// No vertex array exists.
	wipeTex();								// Set all sizes and flags to zero.
	wipeMesh();
	// (m_BMP memory is handled by its own destructor)
}

CquikMeshImg::~CquikMeshImg()
//------------------------------------------------------------------------------
// Destructor (default)
{
	wipeTex();								// destroy contents, release memory
	wipeMesh();
}

void CquikMeshImg::wipeTex(void)
//------------------------------------------------------------------------------
// Wipe out the contents of this object; free any memory it uses, BUT 
// keep any BMP file we may have already loaded.
{
	if(m_pTimg!=NULL)						// ? do we have a texture image?
	{										// Yes... release that memory
		delete [] m_pTimg;					// we reserved in makeTestImage()
		m_pTimg = NULL;
	}
	m_Txsize = m_Tysize = m_Tzsize = 0;		// zero size.
}

void CquikMeshImg::wipeMesh(void)
//------------------------------------------------------------------------------
// Wipe out the contents of this object; free any memory it uses, BUT 
// keep any BMP file we may have already loaded.
{
	if(m_pNorm!=NULL)						// ? do we have a normal vec. array?
	{
		delete [] m_pNorm;					// Yes... release the memory we
		m_pNorm = NULL;						// reserved in makeTestImage()
	}
	if(m_pTexy!=NULL)						// ? do we have texture coord array?
	{
		delete [] m_pTexy;					// Yes... release the memory we
		m_pTexy = NULL;						// reserved in makeTestImage()
	}
	if(m_pVert!=NULL)						// ? do we have a vertex array?
	{
		delete [] m_pVert;					// Yes... release the memory we
		m_pVert = NULL;						// reserved in makeTestImage()
	}
	m_Txsize = m_Tysize = m_Tzsize = 0;		// set image size to zero.
	m_Mxsize = m_Mysize = 0;				// set mesh size to zero.
}

BOOL CquikMeshImg::sizerTex(int Ximg, int Yimg, int Zimg)
//------------------------------------------------------------------------------
// Make our stored image size match (Ximg, Yimg, Zimg), resizing as needed.
// CAREFUL!! Texture map size MUST be a power-of-two!!
{
	// If size given is unchanged or negative or invalid,
	if((m_Txsize==Ximg && m_Tysize==Yimg && m_Tzsize==Zimg) ||
		Ximg<=0 || Yimg<=0 || Zimg<=0 || Zimg>4)
	{
		return(FALSE);						// ignore it. bye.
	}
	wipeTex();								// Wipe previous texmap, if any.
	m_Txsize = Ximg;						// Set recorded size of texture map,
	m_Tysize = Yimg;
	m_Tzsize = Zimg;
	m_pTimg = new GLubyte [Ximg*Yimg*Zimg];	// and allocate new memory.
	return(TRUE);
}

BOOL CquikMeshImg::sizerMesh(int Xmesh, int Ymesh)
//------------------------------------------------------------------------------
// Make our stored mesh size match (Xmesh,Ymesh), resizing as needed.
{
	// If size given is unchanged or negative or invalid,
	if((m_Mxsize==Xmesh && m_Mysize==Ymesh) || Xmesh<=1 || Ymesh<=1)
	{
		return(FALSE);						// ignore it. bye.
	}
	wipeMesh();								// Discard previous contents.
	m_Mxsize = Xmesh;						// Set recorded size of mesh
	m_Mysize = Ymesh;
	m_pNorm = new GLfloat [3*m_Mxsize*m_Mysize];	// for Normal vectors
	m_pTexy = new GLfloat [2*m_Mxsize*m_Mysize];	// for Texture coordinates
	m_pVert = new GLfloat [3*m_Mxsize*m_Mysize];	// for Vertex coordinates
	return(TRUE);							// Report change.
}

/*
void CquikMeshImg::makeTestImgMeshDisplayList(void)			
//------------------------------------------------------------------------------
// Generate a test pattern texture from color ramps in various directions.
// Uses explicit vertex calls; replaced by later version using vertex arrays.
{
int i,j,k;									
float xfrac,yfrac;

	m_Txsize = m_Tysize = 64;					// OpenGL requires power-of-two
												// sized, square texture maps
	m_Tzsize = 4;								// RGBA pixels (transparency!
	m_pTimg = new GLubyte[m_Txsize*m_Tysize*m_Tzsize];	// reserve memory.
	for(j=0,k=0; j<m_Tysize; j++)				
	{											// Make colorful test pattern:
		for(i=0; i<m_Txsize; i++)				// pixel i on scanline j
		{
			m_pTimg[(k + i)*m_Tzsize + 0] =((255-i)*7 +(    j)* 4)%256; //red
			m_pTimg[(k + i)*m_Tzsize + 1] =((255-i)*3 +(255-j)* 8)%256; //green
			m_pTimg[(k + i)*m_Tzsize + 2] =((255-i)*2 +(    j)*11)%256; //blue
			m_pTimg[(k + i)*m_Tzsize + 3] = 170;	// alpha (opacity)
		}
		k += m_Txsize;							// start of scanline 'j'
	}
												// (use default mesh size)
	xfrac = 1.0f/(float)(m_Mxsize-1);			// fractional mesh spacing
	yfrac = 1.0f/(float)(m_Mysize-1);
	//Make a display list for the mesh..........................................
	m_dList = glGenLists(1);				// Reserve a display list for us;
												// m_dList is its identifier#.
	glNewList(m_dList,GL_COMPILE);				// Start the list...
	glBegin(GL_QUADS);							// Make a quadrilateral mesh.
	for(j=0; j<m_Mysize-1; j++)					// height of mesh
	{
		for(i=0; i<m_Mxsize-1; i++)		
		{	// mesh spans 0.0 < x,y < 1.0, but texture coords span -1.0<x,y<2.0
			// so that texture image (0.0<x,y<1.0) is in middle third of the
			// mesh, and GL_TEXTURE_WRAP_S,T settings determine how to draw the
			// outer 2/3 of the image.
			// (Note; MUST call normal,texture,vertex IN THAT ORDER!)
			glNormal3f(0.0f,0.0f,1.0f);					// lower-left
			glTexCoord2f((i  )*xfrac*3.0f - 1.0f,(j  )*yfrac*3.0f - 1.0f);		
			glVertex3f  ((i  )*xfrac,			 (j  )*yfrac, 0.0f);	

			glNormal3f(0.0f,0.0f,1.0f);					// lower-right,
			glTexCoord2f((i+1)*xfrac*3.0f - 1.0f,(j  )*yfrac*3.0f - 1.0f);				
			glVertex3f	((i+1)*xfrac,			 (j  )*yfrac, 0.0f);	
			
			glNormal3f(0.0f,0.0f,1.0f);					// upper-right,
			glTexCoord2f((i+1)*xfrac*3.0f - 1.0f,(j+1)*yfrac*3.0f - 1.0f);				
			glVertex3f	((i+1)*xfrac,			 (j+1)*yfrac,0.0f);	
			
			glNormal3f(0.0f,0.0f,1.0f);					// upper left.
			glTexCoord2f((i  )*xfrac*3.0f - 1.0f,(j+1)*yfrac*3.0f - 1.0f);				
			glVertex3f	((i  )*xfrac,			 (j+1)*yfrac,0.0f);		
		}
	}
	glEnd();		// stop making quads. mesh is finished.
	glEndList();	// stop making display list; finished.
}
*/

void CquikMeshImg::makeTestImgMesh(void)			
//------------------------------------------------------------------------------
// Generate a test pattern texture from color ramps in various directions.
// Builds a vertex array for mesh to allow easy modification & fast rendering.
{
int i,j,k,vnum;									
float xfrac,yfrac;

	m_Txsize = m_Tysize = 64;					// OpenGL requires power-of-two
												// sized, square texture maps
	m_Tzsize = 4;								// RGBA pixels (transparency!
	m_pTimg = new GLubyte[m_Txsize*m_Tysize*m_Tzsize];	// reserve memory.
	for(j=0,k=0; j<m_Tysize; j++)				
	{											// Make colorful test pattern:
		for(i=0; i<m_Txsize; i++)				// pixel i on scanline j
		{
			m_pTimg[(k + i)*m_Tzsize + 0] =((255-i)*7 +(    j)* 4)%256; //red
			m_pTimg[(k + i)*m_Tzsize + 1] =((255-i)*3 +(255-j)* 8)%256; //green
			m_pTimg[(k + i)*m_Tzsize + 2] =((255-i)*2 +(    j)*11)%256; //blue
			m_pTimg[(k + i)*m_Tzsize + 3] = 170;	// alpha (opacity)
		}
		k += m_Txsize;							// start of scanline 'j'
	}
	//Make a vertex array for the mesh.........................................
	m_Mxsize = 23;								// Set a mesh size
	m_Mysize = 17;								// (default is 2,2).
												// Reserve memory for arrays:
	m_pNorm = new GLfloat [3*m_Mxsize*m_Mysize];	// for Normal vectors
	m_pTexy = new GLfloat [2*m_Mxsize*m_Mysize];	// for Texture coordinates
	m_pVert = new GLfloat [3*m_Mxsize*m_Mysize];	// for Vertex coordinates

	xfrac = 1.0f/(float)(m_Mxsize-1);			// fractional mesh spacing
	yfrac = 1.0f/(float)(m_Mysize-1);
												// Fill arrays w/ vertex info:
	for(j=0,k=0; j<m_Mysize; j++, k+=m_Mxsize)	// height of mesh
	{
		for(i=0; i<m_Mxsize; i++)				// width of mesh
		{	// mesh spans 0.0 < x,y < 1.0, but texture coords span -1.0<x,y<2.0
			// so that texture image (0.0<x,y<1.0) is in middle third of the
			// mesh, and GL_TEXTURE_WRAP_S,T settings determine how to draw the
			// outer 2/3 of the image.
			vnum = i+k;							// Vertex number
			m_pNorm[3*vnum+0]= 0.0f;				// Nx,Ny,Nz coords
			m_pNorm[3*vnum+1]= 0.0f; 
			m_pNorm[3*vnum+2]= 1.0f;
												// Tx, Ty coords
			// texture fills center 1/3rd of mesh
		//	m_pTexy[2*vnum+0]= i*xfrac*3.0f - 1.0f;	
		//	m_pTexy[2*vnum+1]= j*yfrac*3.0f - 1.0f;
			// texture fills mesh
				m_pTexy[2*vnum+0]= i*xfrac;		
				m_pTexy[2*vnum+1]= j*yfrac;

			m_pVert[3*vnum+0]= i*xfrac;			// Vx,Vy,Vz coords
			m_pVert[3*vnum+1]= j*yfrac;
			m_pVert[3*vnum+2]= 0.0f;	
/*			ASSERT(i+k < m_Mxsize*m_Mysize);	// Print what's happening
			TRACE("N=%f,%f,%f \t",m_pNorm[i+k],m_pNorm[i+k+1],m_pNorm[i+k+2]);
			TRACE("T=%f,%f \t",m_pTexy[i+k],m_pTexy[i+k+1]);
			TRACE("V=%f,%f%,%f\n",m_pVert[i+k],m_pVert[i+k+1],m_pVert[i+k+2]);
*/		}
	}
}

void CquikMeshImg::drawMe(void)
//------------------------------------------------------------------------------
// Call from within CquikGL::drawOpenGL() only. (Why? fcn assumes lots of openGL
// state info has already been set by CquikGL::OnCreate_GL(), and that we have
// a valid openGL rendering context attached to the Windows device context.
{
int i,j,k;

	glPushAttrib(GL_LIGHTING_BIT);			// save current lighting state
	glPushAttrib(GL_COLOR_BUFFER_BIT);		// save current transparency state
	glPushAttrib(GL_TEXTURE_BIT);			// save current texture state

	glDisable(GL_LIGHTING);					// turn off lights for axes,
	glEnable(GL_BLEND);						// enable tex/poly transparency
	glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
	glColor3f(1.0f,1.0f,1.0f);				// set untextured poly's color to
											// white to avoid modulating the
											// texture map's alpha values.

	//..Set openGL's texture map state params:  (s,t) are texture coords.
	//	WRAP== if texture coords (s,t) fall outside the texture image coords, 
	//	use CLAMP to stay within image, or GL_REPEAT to use image coords 
	// modulo img. size for repeat texture 'tiling'
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP); 
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);
	error_GL("drawOpenGL::tex wrap");
									// Set texture map antialiasing method: 
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
									// when magnifying, use bilinear interpol.
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
									// when reducing, I use bilinear filter
	// Some of the other choices: 
	//  GL_NEAREST, or GL_LINEAR == nearest neighbor, or bilinear filt. src img.
	//  GL_NEAREST_MIPMAP_LINEAR == nearest within 2 mipmaps; blend (default)
	//  GL_LINEAR_MIPMAP_NEAREST == linear interp. within the best mipmap
	//								(has 'resolution popping' in animation)
	//  GL_LINEAR_MIPMAP_LINEAR == trilinear interp. (slowest; looks best)
	//  glTexEnvi(GL_TEXTURE_ENV,GL_TEXTURE_ENV_MODE,GL_DECAL);
	//	glTexEnvi(GL_TEXTURE_ENV,GL_TEXTURE_ENV_MODE,GL_BLEND);
	glTexEnvi(GL_TEXTURE_ENV,GL_TEXTURE_ENV_MODE,GL_MODULATE);
	// Set the way texture is applied to polygon.
	// Choices: 
	//		GL_DECAL (most common). Set poly's on-screen color from texture.
	//			texture alpha (if any) blends with lighted poly color.
	//		GL_MODULATE texture map modulates surface lighting (if no lighting
	//		and polygon color is (1,1,1,1), then texmap sets poly color.
	//		GL_BLEND Fakes using texture as diffuse reflectance for lit poly
	glTexImage2D(GL_TEXTURE_2D,	// Tell openGL which texture map to use;
		0,						// pyramid level to load; 0==base of pyramid);
		m_Tzsize,				// # of color components[1-4]  (3==RGB,4=RGBA)
		m_Txsize,				// width: must be 2^n +2*border 
		m_Tysize,				// height; must be 2^m + 2*border
		0,						// border; must be 0 or 1.
		GL_RGBA,				// pixel format (others include GL_RGB, etc.
		GL_UNSIGNED_BYTE,		// pixel element type.
		m_pTimg);				// ptr to the texture image to load.
	error_GL("drawOpenGL--TexImage2D");	// Report any mistakes we made
	glEnable(GL_TEXTURE_2D);		// Last of all, enable 2D texture mapping.
	error_GL("drawOpenGL::tex 2D");

	//..Make a lit rectangle we'll texture map......
	glPolygonMode(GL_FRONT, GL_FILL);		// Solid, front-facing polys,
	//	glDisable(GL_LIGHTING);					// without lighting complications,
	
	// Use vertex-describing arrays to render the textured mesh:
	glMatrixMode(GL_MODELVIEW);					// save current modelview...
	glPushMatrix();	
	glTranslatef(-0.5f,-0.5f,-1.0f);			// move to P2 image position
												// (Draw a mesh of X,Y verts 
												// using X-1,Y-1 quadrilaterals)
											
						// makeTestImgMesh() put all vertex data into arrays at
						// pointers. Tell GL how to use them:
	glEnableClientState(GL_NORMAL_ARRAY);		// Enable arrays of vertex data
	glEnableClientState(GL_TEXTURE_COORD_ARRAY);
	glEnableClientState(GL_VERTEX_ARRAY);
												// Tell GL how to use them:
	glNormalPointer(    GL_FLOAT,0,m_pNorm);	// 3-vector,float,no stride
	glTexCoordPointer(2,GL_FLOAT,0,m_pTexy);	// 2-vector,float,no stride
	glVertexPointer(  3,GL_FLOAT,0,m_pVert);	// 3-vector,float,no stride.

	glBegin(GL_QUADS);
	for(j=0,k=0; j<m_Mysize-1; j++, k+=m_Mxsize)// (k is vert# for column 0)
	{
		for(i=0; i<m_Mxsize-1; i++)
		{
			glArrayElement(i+k);				// lower left corner of a quad
			glArrayElement(i+k+1);				// lower right,
			glArrayElement(i+k+1+m_Mxsize);		// upper right
			glArrayElement(i+k  +m_Mxsize);		// upper left.
		}
	}
	glEnd();									// Done with the mes.
	glPopMatrix();						// restore original MODELVIEW matrix.

	glPopAttrib();						// restore original texturing state
	glPopAttrib();						// restore original transparency state
	glPopAttrib();						// restore original lighting state
}

void CquikMeshImg::error_GL(char *label)
//------------------------------------------------------------------------------
// If OpenGL had any errors, print string found at 'label' to report where it
// happened in our code, and then print the error report. Otherwise do nothing.
// (NOTE: the 'TRACE()' macro works just like the 'printf()' function, EXCEPT 
//    output appears at bottom of the VisualC++6.0 window on the 'debug' tab.)
{
GLenum errorno;								// OpenGL error code.

	errorno = glGetError();					
	if (errorno != GL_NO_ERROR)				// was there any unreported error?
	{										// YES: tell me about it.
		TRACE("%s had error: #(%d) %s\r\n", 
		label, errorno, gluErrorString(errorno)); // (give OpenGL's explanation)
	}
}

void CquikMeshImg::meshWrinkle(float zoffset)
//------------------------------------------------------------------------------
// Displace odd-numbered vertices in the +z direction by 'zoffset'. Not very
// useful, but demonstrates how to move mesh vertices within vertex array, and
// shows vertex locations.
{
	
}

BOOL CquikMeshImg::fileLoad(OPENFILENAME *pName)
//------------------------------------------------------------------------------
// DON'T CALL THIS FUNCTION YOURSELF! the CquikGL class calls it in response
// to user input.  Loads texture map image from Windows BMP file <*pName>.bmp
// if it exists, and also loads mesh geometry from <*pName.bmp>.  
{
BOOL isOK;
long int siz,n,i,j,imax,jmax,k;
BYTE *pSrc;

	isOK = m_BMP.fileRead(pName);			// Open, read, and close a BMP file.
	if(isOK) wipeTex();						// discard prev. texture map.
	//......................................Find the best size for texture map.
	// OpenGL texture map images must have powers-of-two sizes in memory. Odd-
	// sized BMP files make larger texture maps sized 2^n x 2^n (integer n>0),
	// and are read in starting at pixel 0,0; and I scale texture coords to use 
	// only the image-filled part.  (Unused texture pixels are opaque white).
	m_Txsize = m_BMP.getXsize() -1;			// Image's largest pixel address
	m_Tysize = m_BMP.getYsize() -1;			
	siz = 1;								// Find texture map size needed;
	for(n=0; m_Txsize>0 || m_Tysize>0; n++)	// (find image size: siz==2^n)
	{
		m_Txsize = m_Txsize >> 1;			// arithmetic shift-right ( /2 )
		m_Tysize = m_Tysize >> 1;
		siz = siz << 1;						// compute 2^n.
	}										// Use RGBA pixels (not RGB).
	sizerTex(siz,siz,4);					// Set the texture map size.
	if(	siz != m_BMP.getXsize() ||			// If the BMP file isn't a
		siz != m_BMP.getYsize())			// perfect fit in the texture map,
	{
		memset(m_pTimg,0xff,m_Txsize*m_Tysize*m_Tzsize);	
											// Fill texture map with 255's so
	}										// so unfilled part is opaque white.
	//......................................Fill the texture map from BMP file.
	imax = m_BMP.getXsize();
	jmax = m_BMP.getYsize();
	ASSERT(m_Tzsize==4);					// (be sure texture is RGBA)
	for(j=0,k=0; j<jmax; j++)				// For each scanline of BMP file
	{
		pSrc = m_BMP.getPtr(0,j);			// Get ptr to 1st byte of scanline,
		for(i=0; i<imax; i++)				// then copy pixel values:
		{
			m_pTimg[i*m_Tzsize + k+0] = pSrc[i*3 +2];	// red,
			m_pTimg[i*m_Tzsize + k+1] = pSrc[i*3 +1];	// green,
			m_pTimg[i*m_Tzsize + k+2] = pSrc[i*3 +0];	// blue. 
			m_pTimg[i*m_Tzsize + k+3] = 128;			// Alpha--(opacity)
			//(Note Microsoft does it backwards--stores B,G,R pixels, not RGB.)
		}		
		k += m_Txsize * m_Tzsize;			// go to next texture map scanline
	}
	return(TRUE);								// Report Success.
}

BOOL CquikMeshImg::fileSave(OPENFILENAME *pName)
//------------------------------------------------------------------------------
// DON'T CALL THIS FUNCTION YOURSELF! the CquikGL class calls it in response
// to user input.  Saves texture map image to Windows BMP file <*pName>.bmp
// and also loads mesh geometry from <pName>bmp.
{
	return(m_BMP.fileWrite(pName));			// Save the current m_BMP image.
}

//=======================
//=======================
//=======================
//
//	CquikBMP Class
//
//=======================
//=======================
//=======================


CquikBMP::CquikBMP()
//------------------------------------------------------------------------------
// Default constructor/destructor
{
	m_pImg = NULL;							// NULL means no memory reserved.
	m_xsize = m_ysize = m_linesize = 0;
	m_BMPone.bfType = 0;					// ="BM"=0x4D42 for Windows BMP file		
	m_BMPone.bfSize = 0;					// total file size in bytes.
	m_BMPone.bfReserved1 = 0;				// Always zero.
	m_BMPone.bfReserved2 = 0;				// Always zero.
	m_BMPone.bfOffBits = 0;					// #bytes in m_BMPone and m_BMPtwo.
	m_pBMPtwo = NULL;						// NULL means no memory reserved.
}

CquikBMP::~CquikBMP()
//------------------------------------------------------------------------------
// Default constructor/destructor
{
	wipe();									// release any memory allocated.
}

void CquikBMP::wipe()
//------------------------------------------------------------------------------
// Discard any current contents.
{
	// zero out the two structs used at the start of each BMP file:...........
	m_BMPone.bfType = 0x4D42;				// ASCII "BM" for Windows BMP file		
	m_BMPone.bfSize = 0;					// total file size in bytes.
	m_BMPone.bfReserved1 = 0;				// Always zero.
	m_BMPone.bfReserved2 = 0;				// Always zero.
	m_BMPone.bfOffBits = 0;					// #bytes in m_BMPone and m_BMPtwo.
	if(m_pBMPtwo != NULL)					// Do we have one in memory?
	{
		delete [] m_pBMPtwo;				// Yes... release the memory we
		m_pBMPtwo = NULL;					// reserved in sizer or fileRead().
	}
	m_xsize = m_ysize = m_linesize = 0;		// Set size to zero.
	if(m_pImg != NULL)						// do we have an image buffer?
	{
		delete [] m_pImg;					// Yes... release memory from
		m_pImg = NULL;						// sizer() or fileRead().
	}
}

BOOL CquikBMP::sizer(int Xsize, int Ysize)
//------------------------------------------------------------------------------
// Change the image buffer size if needed; any size change will discard current 
// buffer contents and re-set all the BMP file info kept with it.
{
long int siz;
											// Do we need a new image size?
	if((Xsize == m_xsize && Ysize == m_ysize) || Xsize<=0 || Ysize<=0)
	{										// No;  change nothing.
		return(FALSE);						// report error.
	}										// YES......change everything:
	m_xsize = Xsize;						// Set image size in pixels;
	m_ysize = Ysize;
	m_linesize = Xsize*3;					// Find # of bytes per scanline
	m_linesize += (4 - m_linesize%4)%4;		// (add 'padding' bytes as needed)
			// !!!CAREFUL!!! the BMP files use an extra 0-3 bytes at the end of 
			// each scanline to ensure lengths are always a multiple of 4.
			// (required for Window's 32-bit memory access scheme).
	if(m_pImg != NULL)						// Does an image buffer exist?
	{
		delete [] m_pImg;					// Yes-- release its memory.
		m_pImg = NULL;
	}
	siz = (long int)m_linesize * (long int)m_ysize;	
											// find image buffer size.
	m_pImg = new BYTE [siz];				// grab memory for it.
			// Set values of both BMP structs................................... 
	if(m_pBMPtwo!=NULL)						// Does one already exist?
	{
		delete [] m_pBMPtwo;				// YES--release its memory.
		m_pBMPtwo = NULL;
	}										// and reserve space for a new one.
	m_pBMPtwo = (BITMAPINFO *) new BYTE [sizeof(BITMAPINFO)]; 
	// m_BMPtwo has just two members:
				// m_BMPtwo.bmiHeader==one Windows BITMAPINFOHEADER struct.
				//						that holds many fixed-sized constants.
				// m_BMPtwo.bmiColors[]==one or more Windows RGBQUAD structs.
				//						in an array used as a color table or 
				//						'palette' to convert indexed pixels 
				//						(e.g. 1,4,8,16 bits) to RGB screen
				//						values of 24 bits.  Ignored if BMP file
				//						holds 24 bit values in the first place..
	m_pBMPtwo->bmiHeader.biWidth = m_xsize;	// Image size, in **pixels**
	m_pBMPtwo->bmiHeader.biHeight= m_ysize;
	m_pBMPtwo->bmiHeader.biPlanes		=1;	// Vestigial. Always =1. 
	m_pBMPtwo->bmiHeader.biBitCount	   =24;	// =24 for true-color. [or 1,4,8]
	m_pBMPtwo->bmiHeader.biCompression	=0;	// =0 for no compression. [or 1,2]
	m_pBMPtwo->bmiHeader.biSizeImage = siz;	// padded bitmap size in bytes. 
	m_pBMPtwo->bmiHeader.biXPelsPerMeter=0;	// onscreen size: Usually just 0.
	m_pBMPtwo->bmiHeader.biYPelsPerMeter=0;	// "		"		"		"
	m_pBMPtwo->bmiHeader.biClrUsed		=0;	// Palette size & dither priorities; 
	m_pBMPtwo->bmiHeader.biClrImportant	=0;	// (set to zero for true-color)
	//
	m_pBMPtwo->bmiColors[0].rgbBlue		=0;	// Set colors for the minimal
	m_pBMPtwo->bmiColors[0].rgbGreen	=0;	// palette: note each color-table
	m_pBMPtwo->bmiColors[0].rgbRed		=0;	// entry is 4 bytes. 
	m_pBMPtwo->bmiColors[0].rgbReserved	=0;	// 
	m_pBMPtwo->bmiHeader.biSize = 
		sizeof(m_pBMPtwo->bmiColors) + sizeof(m_pBMPtwo->bmiHeader);	
	// size of bmiHeader member in bytes (NOTE! bmiHeader size doesn't vary--
	// but this lets MS extend the format) However, bmiColors[] array size DOES 
	// vary: min size=4 bytes, even for 24-bit BMP files that ignore palettes. 
	// If you add palettes later, they must be included in bmiHeader.biSize.
	// And now the FIRST struct found in a BMP file:............................
	m_BMPone.bfType = 0x4D42;				// the ASCII letters "BM"
	m_BMPone.bfSize = sizeof(BITMAPFILEHEADER) + sizeof(BITMAPINFO) + siz;	
	// (Should be the file size in bytes, but m_BMPone.bfSize is an UNRELIABLE 
	// number in some BMP files due to errors in early Windows specs. Ignore it.
	m_BMPone.bfReserved1=0;					// Always zero.
	m_BMPone.bfReserved2=0;					// Always zero.
	m_BMPone.bfOffBits= sizeof(BITMAPFILEHEADER) + sizeof(BITMAPINFO);
											// Bytes between the start of file
											// and the first bitmap byte.
	return(TRUE);
}

BOOL CquikBMP::fileRead(OPENFILENAME *pName)
//------------------------------------------------------------------------------
// Open, read, & close named file. (Discards any existing contents)
{
CFile Fimg;								// Image file object.
CFileException errImg;					// Records any file system errors.
BYTE *pBuf;								// ptr to start of BMP scanline.
BOOL isOK;
DWORD siz;
int xs,ys,j;

										// Use Windows file opening method:
	isOK = Fimg.Open(pName->lpstrFile,	CFile::modeRead | 
										CFile::shareDenyWrite | 
										CFile::typeBinary,	&errImg);
	if(isOK==FALSE) 
	{
		TRACE("Can't open file %s: err= %u\n",pName->lpstrFile,errImg.m_cause);
		return FALSE;
	}
	//..................................Read in the first struct from BMP file:
	Fimg.Read((void *)&m_BMPone,sizeof(m_BMPone));
	if( m_BMPone.bfType != 0x4D42 ||	// ="BM"=0x4D42 for Windows BMP file 
		m_BMPone.bfReserved1 != 0 ||	// and Reserved1 & 2 should be zero.
		m_BMPone.bfReserved2 != 0)			
	{
		TRACE("file %s doesn't hold Windows BMP info\n.");
		Fimg.Close();					// close the file & quit.
		return(FALSE);
	}
	//..................................Read in the second struct from BMP file;
	if(m_pBMPtwo != NULL)					// Does it already exist?
	{										// (e.g. from previous file)
		delete [] m_pBMPtwo;				// YES. release the memory it used.
		m_pBMPtwo = NULL;
	}
	siz = m_BMPone.bfOffBits - sizeof(m_BMPone);// Find 2nd struct's size:
	m_pBMPtwo = (BITMAPINFO *)new BYTE [siz];	// get enough memory for it,
	Fimg.Read((void *)m_pBMPtwo,siz);			// and read it from the file.
	if(	m_pBMPtwo->bmiHeader.biPlanes!=1	  ||
		m_pBMPtwo->bmiHeader.biBitCount!=24	  ||
		m_pBMPtwo->bmiHeader.biCompression!=0)	// Uncompressed,true-color only!
	{
		TRACE("File %s:Sorry; uncompressed 24-bit only.\n",Fimg.GetFileTitle());
		wipe();								// discard all current contents.
		return(FALSE);
	}
	xs = m_pBMPtwo->bmiHeader.biWidth;		// get the image size in pixels;
	ys = m_pBMPtwo->bmiHeader.biHeight;
	//......................................Fill the image buffer from the file.
	sizer(xs,ys);							// Set size properly (NOTE! replaces
											// BMPone,BMPtwo struct contents.
	for(j=0; j<ys; j++)						// For each BMP file scanline,
	{
		pBuf = getPtr(0,j);					// ptr to start of scanline,
		Fimg.Read((void *)pBuf, m_linesize);// read RGB values from file to it.
	}
	//.......................................Orderly exit: 
	Fimg.Close();								// Close the file.
	return(TRUE);								// Report Success.
}

BOOL CquikBMP::fileWrite(OPENFILENAME *pName)
//------------------------------------------------------------------------------
// Open, write, & close named file using existing buffer contents.
{
CFile Fimg;								// Image file object.
CFileException errImg;					// Records any file system errors.
BYTE *pBuf;								// ptr to start of BMP scanline.
BOOL isOK;
DWORD siz;
int xs,ys,j;

	if(m_pImg==NULL || m_pBMPtwo==NULL || m_xsize<=0 || m_ysize<=0)
	{
		return(FALSE);							// invalid image; can't write.
	}
										// Use Windows file opening method:
	isOK = Fimg.Open(pName->lpstrFile,CFile::modeWrite | 
										CFile::modeCreate | 
										//CFile::modeNoTruncate |
										CFile::typeBinary,	&errImg);
	if(isOK==FALSE) 
	{
		TRACE("Can't create file %s: err= %u\n",pName->lpstrFile,errImg.m_cause);
		Fimg.Close();
		return FALSE;
	}
	//.................................Write out the first struct from BMP file:
	ASSERT( m_BMPone.bfType == 0x4D42 || // ="BM"=0x4D42 for Windows BMP file 
			m_BMPone.bfReserved1 == 0 ||// and Reserved1 & 2 should be zero.
			m_BMPone.bfReserved2 == 0);			
	Fimg.Write(&m_BMPone,sizeof(m_BMPone));
	//................................Write out the second struct from BMP file;
	ASSERT( m_pBMPtwo != NULL);					// (2nd struct doesn't exist?!?)
	siz = m_BMPone.bfOffBits - sizeof(m_BMPone);// Find 2nd struct's size:
	Fimg.Write( m_pBMPtwo, siz);				// and read it from the file.
	if(	m_pBMPtwo->bmiHeader.biPlanes!=1	  ||
		m_pBMPtwo->bmiHeader.biBitCount!=24	  ||
		m_pBMPtwo->bmiHeader.biCompression!=0)	// Uncompressed,true-color only!
	{
		TRACE("File %s:Sorry; uncompressed 24-bit only.\n",Fimg.GetFileTitle());
		wipe();								// discard all current contents.
		Fimg.Close();						// close the file,
		return(FALSE);						// and quit.
	}
	xs = m_pBMPtwo->bmiHeader.biWidth;		// get the image size in pixels;
	ys = m_pBMPtwo->bmiHeader.biHeight;
	//......................................Write the image buffer to the file.
	for(j=0; j<ys; j++)						// For each BMP file scanline,
	{
		pBuf = getPtr(0,j);					// ptr to start of scanline,
		Fimg.Write(pBuf, m_linesize);		// write RGB values to file to it.
	}
	//.......................................Orderly exit: 
	Fimg.Close();								// Close the file.
	return(TRUE);
}

BYTE CquikBMP::getPixel(int x, int y, int z)
//------------------------------------------------------------------------------
// Read image buffer; z=0 for red, z=1 for green, z=2 for blue.
{
	return(m_pImg[x*3 + y*m_linesize + z]);
}

BYTE CquikBMP::putPixel(int x, int y, int z, BYTE val)
//------------------------------------------------------------------------------
// Set an image buffer value
{
	m_pImg[x*3 + y*m_linesize + z] = val;
	return(TRUE);
}

BYTE* CquikBMP::getPtr(int x, int y)
//------------------------------------------------------------------------------
// Return pointer to image buffer at location x,y.
{
	return( m_pImg + x*3 + y*m_linesize );	// x*3 for RGB triplets
}

